ROTATED_PROXY_ENABLED = True
PROXY_STORAGE = 'scrapy_rotated_proxy.extensions.file_storage.FileProxyStorage'
PROXY_FILE_PATH = ''
# PROXY_STORAGE = 'scrapy_rotated_proxy.extensions.mongodb_storage.MongoDBProxyStorage'
PROXY_MONGODB_HOST = '127.0.0.1'
PROXY_MONGODB_PORT = 27017
PROXY_MONGODB_USERNAME = None
PROXY_MONGODB_PASSWORD = None
PROXY_MONGODB_AUTH_DB = 'admin'
PROXY_MONGODB_DB = 'proxy_management'
PROXY_MONGODB_COLL = 'proxy'
PROXY_MONGODB_COLL_INDEX = []
PROXY_SLEEP_INTERVAL = 60*60*24
PROXY_SPIDER_CLOSE_WHEN_NO_PROXY = True
PROXY_RELOAD_ENABLED = False
